# -*- coding: utf-8 -*-
# Copyright (C) 2020 fjord-technologies
# SPDX-License-Identifier: GPL-3.0-or-later
"""xtrargon_ng.modules.quorum"""

import re
import logging
import requests

from dwho.classes.modules import DWhoModuleBase, MODULES
from httpdis.ext.httpdis_json import HttpReqErrJson
from sonicprobe.libs import network, urisup, xys
from sonicprobe.libs.moresynchro import RWLock

from ..classes.abstract.db import XtrargonDBBase


LOG                        = logging.getLogger('xtrargon_ng.modules.quorum')

DB_NODE_LETTERS            = {'a': 'b',
                              'b': 'a'}

DEFAULT_CNF_FILE           = "quorum.cnf"

DEFAULT_BACKUPS_STATUS_URI = "http://:0/backups/status"
DEFAULT_FORBID_VIP_URI     = "http://:0/server/forbid-vip"
DEFAULT_REQUEST_TIMEOUT    = 15

MATCH_DB_NODE              = re.compile(r'^([a-z0-9][a-z0-9\-\.]+[a-z0-9]-[0-9]+)([a-z])(\.[a-z0-9\-\.]+[a-z0-9]|)$').match

xys.add_callback('quorum.host', network.valid_domain)

class QuorumDB(XtrargonDBBase):
    pass

class QuorumModule(DWhoModuleBase):
    MODULE_NAME = 'quorum'

    LOCK = RWLock()

    # pylint: disable-msg=attribute-defined-outside-init
    def safe_init(self, options):
        self.lock_timeout = self.config['general']['lock_timeout']
        self.request_timeout = DEFAULT_REQUEST_TIMEOUT

        backups_status_uri = DEFAULT_BACKUPS_STATUS_URI
        forbid_vip_uri = DEFAULT_FORBID_VIP_URI

        if isinstance(self.config.get('quorum'), dict):
            if self.config['quorum'].get('backups_status_uri'):
                backups_status_uri = self.config['quorum']['backups_status_uri']

            if self.config['quorum'].get('forbid_vip_uri'):
                forbid_vip_uri = self.config['quorum']['forbid_vip_uri']

            if 'request_timeout' in self.config['quorum']:
                self.request_timeout = float(self.config['quorum']['request_timeout'])

        self.backups_status_uri = list(urisup.uri_help_split(backups_status_uri))
        self.forbid_vip_uri = list(urisup.uri_help_split(forbid_vip_uri))

    @staticmethod
    def _get_neighbor_node(host):
        node = host.lower()

        m = MATCH_DB_NODE(host)
        if not m:
            raise HttpReqErrJson(415, "invalid node host argument: %r" % node)

        letter = m.group(2)
        if letter not in DB_NODE_LETTERS:
            raise HttpReqErrJson(415, "invalid node host letter: %r" % node)

        host_splitted    = list(m.groups())
        host_splitted[2] = DB_NODE_LETTERS[letter]

        return ''.join(host_splitted)

    def _build_uri(self, uri, host = None):
        uri = list(uri)
        if uri[1]:
            uri[1] = list(uri[1])
        else:
            uri[1] = [None, None, None, None]

        if host and not uri[1][2]:
            uri[1][2] = host

        if uri[1][3] == '0':
            uri[1][3] = self.config['general'].get('listen_port')

        return urisup.uri_help_unsplit(uri)

    BACKUPS_STATUS_QSCHEMA = xys.load("""
    host: !~~callback(quorum.host)
    """)

    def backups_status(self, request):
        params = request.query_params()

        if not isinstance(params, dict):
            raise HttpReqErrJson(400, "invalid arguments type")

        if not xys.validate(params, self.BACKUPS_STATUS_QSCHEMA):
            raise HttpReqErrJson(415, "invalid arguments for command")

        host = self._get_neighbor_node(params['host'])

        if not self.LOCK.acquire_read(self.lock_timeout):
            raise HttpReqErrJson(503, "unable to take LOCK for reading after %s seconds" % self.lock_timeout)

        req = None

        try:
            LOG.info("get backups status on %r", host)

            req = requests.get(self._build_uri(self.backups_status_uri, host),
                               timeout = self.request_timeout)
            if not req.text:
                return None

            try:
                return req.json()
            except ValueError:
                req.raise_for_status()
        except HttpReqErrJson:
            raise
        except Exception as e:
            LOG.exception(e)
            raise HttpReqErrJson(503, repr(e))
        finally:
            if req:
                req.close()
            self.LOCK.release()


    FORBID_VIP_STATUS_QSCHEMA = xys.load("""
    host: !~~callback(quorum.host)
    """)

    def forbid_vip_status(self, request):
        params = request.query_params()

        if not isinstance(params, dict):
            raise HttpReqErrJson(400, "invalid arguments type")

        if not xys.validate(params, self.FORBID_VIP_STATUS_QSCHEMA):
            raise HttpReqErrJson(415, "invalid arguments for command")

        host = self._get_neighbor_node(params['host'])

        if not self.LOCK.acquire_read(self.lock_timeout):
            raise HttpReqErrJson(503, "unable to take LOCK for reading after %s seconds" % self.lock_timeout)

        req = None

        try:
            LOG.info("get forbid vip status on %r", host)

            req = requests.get(self._build_uri(self.forbid_vip_uri, host),
                               timeout = self.request_timeout)
            if not req.text:
                return None

            try:
                return req.json()
            except ValueError:
                req.raise_for_status()
        except HttpReqErrJson:
            raise
        except Exception as e:
            LOG.exception(e)
            raise HttpReqErrJson(503, repr(e))
        finally:
            if req:
                req.close()
            self.LOCK.release()


    REPLICA_STATUS_QSCHEMA = xys.load("""
    host: !~~callback(quorum.host)
    """)

    def replica_status(self, request):
        params = request.query_params()

        if not isinstance(params, dict):
            raise HttpReqErrJson(400, "invalid arguments type")

        if not xys.validate(params, self.REPLICA_STATUS_QSCHEMA):
            raise HttpReqErrJson(415, "invalid arguments for command")

        host = self._get_neighbor_node(params['host'])

        if not self.LOCK.acquire_read(self.lock_timeout):
            raise HttpReqErrJson(503, "unable to take LOCK for reading after %s seconds" % self.lock_timeout)

        qdb = None

        try:
            LOG.info("get replica status on %r", host)
            qdb = QuorumDB()
            qdb.init_db(module = self, default_cnf_file = DEFAULT_CNF_FILE, host = host)
            qdb.create_connection()

            return qdb.get_slave_status()
        except HttpReqErrJson:
            raise
        except Exception as e:
            LOG.exception(e)
            raise HttpReqErrJson(503, repr(e))
        finally:
            if qdb:
                qdb.close_connection()
            self.LOCK.release()


if __name__ != "__main__":
    def _start():
        MODULES.register(QuorumModule())
    _start()
